
/* bison input file */
%{
/******************************************************************************
    Copyright (c) 1996-2000 Synopsys, Inc.    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/tapinprogram.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/
#include <stdio.h>
#include "mymalloc.h"
#include "group_enum.h"
#include "LibHead.h"
#include "LibParse.h"
#include "LibGroup.h"
#include "LibEnumMap.h"
	
  static int gsindex = 0;

  static LibGroup* gs[1000];
  static LibErrorT   err;
  static LibAttr* curr_attr;
  void push_group(LibHead *h);
  void pop_group();
  bool supported(const char* nm, bool isAttr);
  void deleteParams(LibValEList* paramlist);
  LibValType convert_vt(char *type);
  int lineno;
  int syntax_errors;
  static char PB[8000]; /* so as not to have a bunch of local buffers */
  
  extern void liberty_delete_lex_buffer();
  extern void liberty_flush_string_buf();

  void yyerror(const char *s);
  int yylex();
  void make_complex(LibHead *h);
  void make_simple(char *name, LibVal *v);
 
#define LIBERTY_PARSER_ABORT {liberty_delete_lex_buffer(); YYABORT;}
 
%}

%union
{
  char *str;
  double num;
  LibGroup *group;
  LibAttr *attr;
  LibVal *val;
  LibDefine *def;
  LibHead *head;	
  LibValEList *vlist;
}


%token COMMA SEMI LPAR RPAR LCURLY RCURLY COLON KW_DEFINE KW_DEFINE_GROUP KW_TRUE KW_FALSE

%token <num> NUM
%token <str> STRING IDENT HIER_IDENT

%type <group>     group file statements statement
%type <def>       define define_group
%type <vlist>     param_list
%type <val>       attr_val
%type <str>	  s_or_i
%type <str>	  attr_s_or_i
%type <head>      head attr_head
%%

file            : {
                    gs[gsindex++] = LibParse::group();
		    lineno = 1;
		    syntax_errors= 0;
                  }
                  group
		  {
		    liberty_flush_string_buf();
		    pop_group();
		  }
                ;

group           : head LCURLY
                  {
		    push_group($1);
		    delete $1;
		  }
                  statements RCURLY
                  {
		    pop_group();
		  }
                | head LCURLY
                  {
		    push_group($1);
		    delete $1;
		  }
                  RCURLY
                  {
		    pop_group();
		  }
                | head error
                  {
                    delete $1;
                    LIBERTY_PARSER_ABORT;
                  }
                | head LCURLY error
                  {
                    delete $1;                   
                    LIBERTY_PARSER_ABORT;
                  }
                ;


statements 	: statement {}
                | statements statement  {}
		;


statement 	: simple_attr {}
		| complex_attr {}
		| define {}
		| define_group {}
	       	| group  {}
                | SEMI {}
                | error
                  {
                    LIBERTY_PARSER_ABORT;
                  }
		;

simple_attr	: IDENT COLON attr_val
                  {
		    make_simple($1,$3);
		    my_free($1);
		  }
                  SEMI
		| IDENT COLON attr_val
                  {
		    make_simple($1,$3);
		    my_free($1);
		  }
                | IDENT error
                  {
                    my_free($1);
                    LIBERTY_PARSER_ABORT;
                  }
                | IDENT COLON error
                  {
                    my_free($1);
                    LIBERTY_PARSER_ABORT;
                  }
		;

complex_attr 	: attr_head  SEMI
                  {
		    make_complex($1);
		    delete $1;
		  }
                | attr_head
                  {
		    make_complex($1);
		    delete $1;
		  }
		;

attr_head        : IDENT LPAR param_list RPAR
                  {
		    if (supported($1, true)) {
		      LibFileInfo f(lineno);
		      UString name($1);
		      $$ = new IN_LIBERTY_POOL LibHead(name, f, $3);
		    } else {
		      deleteParams($3);
		      delete $3;
		      $$ = NULL;
		    }
		    my_free($1);
		  }
                | IDENT LPAR RPAR
                  {
		    if (supported($1, true)) {
		      LibFileInfo f(lineno);
		      UString name($1);
		      $$ = new IN_LIBERTY_POOL LibHead(name, f);
		    } else {
		      $$ = NULL;
		    }
		    my_free($1);
		  }
                | IDENT LPAR error
                  {
                    my_free ($1);
                    LIBERTY_PARSER_ABORT;
                  }
                | IDENT LPAR param_list error
                  {
		    deleteParams($3);
		    delete $3;
		    my_free ($1);
                    LIBERTY_PARSER_ABORT;
                  }  
                ;

head            : IDENT LPAR param_list RPAR
                  {
		    if (supported($1, false)) {
		      LibFileInfo f(lineno);
		      UString name($1);
		      $$ = new IN_LIBERTY_POOL LibHead(name, f, $3);
		    } else {
		      deleteParams($3);
		      delete $3;
		      $$ = NULL;
		    }
		    my_free($1);
		  }
                | IDENT LPAR RPAR
                  {
		    if (supported($1, false)) {
		      LibFileInfo f(lineno);
		      UString name($1);
		      $$ = new IN_LIBERTY_POOL LibHead(name, f);
		    } else {
		      $$ = NULL;
		    }
		    my_free($1);
		  }
                | IDENT LPAR error
                  {
                    my_free ($1);
                    LIBERTY_PARSER_ABORT;
                  }
                | IDENT LPAR param_list error
                  {
		    deleteParams($3);
		    delete $3;
		    my_free ($1);
                    LIBERTY_PARSER_ABORT;
                  }  
                ;

param_list      : attr_val
                  {
		    $$= new IN_LIBERTY_POOL LibValEList;
		    $$->append($1);
		  }
                | param_list COMMA attr_val
                  {
		    $$ = $1;
		    $$->append($3);
		  }
                | param_list attr_val
                  {
		    $$ = $1;
		    $$->append($2);
		  }
                | param_list COMMA error
                  {
		    deleteParams($1);
		    delete $1;
                    LIBERTY_PARSER_ABORT;
                  }
                ;

define          : KW_DEFINE LPAR s_or_i COMMA s_or_i COMMA s_or_i RPAR SEMI  
		  {
		    if (gs[gsindex-1]) {
		      LibFileInfo f(lineno);
		      gs[gsindex-1]->createDefine($3, $5, convert_vt($7), f, err);
		    }
		    my_free($3);
		    my_free($5);
		    my_free($7);
		  }
                | KW_DEFINE LPAR s_or_i error
                  {
                    my_free($3);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE LPAR s_or_i COMMA error
                  {
                    my_free($3);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE LPAR s_or_i COMMA s_or_i error
                  {
                    my_free($3);
                    my_free($5);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE LPAR s_or_i COMMA s_or_i COMMA error
                  {
                    my_free($3);
                    my_free($5);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE LPAR s_or_i COMMA s_or_i COMMA s_or_i error
                  {
                    my_free($3);
                    my_free($5);
                    my_free($7);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE LPAR s_or_i COMMA s_or_i COMMA s_or_i RPAR error
                  {
                    my_free($3);
                    my_free($5);
                    my_free($7);
                    LIBERTY_PARSER_ABORT;
                  }
		;

define_group    : KW_DEFINE_GROUP LPAR s_or_i COMMA s_or_i RPAR SEMI
                  {
		    if (gs[gsindex-1]) {
		      LibFileInfo f(lineno);
		      gs[gsindex-1]->createDefine($3, $5,
						  LIBERTY__VAL_UNDEFINED, f, err);
		    }
		    my_free($3);
		    my_free($5);
		  }
                | KW_DEFINE_GROUP LPAR s_or_i error
                  {
                    my_free($3);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE_GROUP LPAR s_or_i COMMA error
                  {
                    my_free($3);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE_GROUP LPAR s_or_i COMMA s_or_i error
                  {
                    my_free($3);
                    my_free($5);
                    LIBERTY_PARSER_ABORT;
                  }
                | KW_DEFINE_GROUP LPAR s_or_i COMMA s_or_i RPAR error
                  {
                    my_free($3);
                    my_free($5);
                    LIBERTY_PARSER_ABORT;
                  }
		;

s_or_i          : STRING {$$ = $1;}
		| IDENT {$$ = $1;}
		;

attr_s_or_i     : s_or_i {$$ = $1;}
                | HIER_IDENT {$$ = $1;}
                ;
                 
attr_val        : NUM
                  {
                    $$ = new IN_LIBERTY_POOL LibVal(LIBERTY__VAL_DOUBLE);
		    $$->double_val($1);
		  }
                | attr_s_or_i
                  {
		    $$ = new IN_LIBERTY_POOL LibVal(LIBERTY__VAL_STRING);
		    $$->string_val($1);
		  }
                | KW_TRUE
                  {
		    $$= new IN_LIBERTY_POOL LibVal(LIBERTY__VAL_BOOLEAN);
		    $$->bool_val(true);
		  }
                | KW_FALSE
                  {
		    $$= new IN_LIBERTY_POOL LibVal(LIBERTY__VAL_BOOLEAN);
		    $$->bool_val(false);
		  }
                ;
%%

void
push_group(LibHead *h)
{
  LibVal *v;
  //  LibErrorT err;
  LibMsgHandlerT MsgPrinter;
  group_enum ge;

  LibParse::control()->interrupted();

  MsgPrinter = LibParse::msgHandler(); /* the printer is in another file! */

  if (!h || !gs[gsindex-1]) {
    gs[gsindex++] = NULL;
    return;
  }
  gs[gsindex] = gs[gsindex-1]->createGroup(0,h->name(),h->fileInfo(),err);
  ge = (group_enum)gs[gsindex]->typeID();
  
  if (err == LIB_OBJECT_ALREADY_EXISTS &&
      (ge != LIBERTY_GROUPENUM_internal_power))	{
    sprintf(PB,"%s:%d: The group name %s is already used in this context.",
	    LibParse::filename(), lineno, (str_t)(h->name()));
    (*MsgPrinter)(LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, PB);
  }
  if (err == LIB_INVALID_NAME) {
    sprintf(PB,"%s:%d: The group name \"%s\" is invalid.",
	    LibParse::filename(), lineno, (str_t)(h->name()));
    (*MsgPrinter)(LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, PB);
  }
  gsindex++;
	
  LibValIter viter = h->valIter();
  while ((v = h->nextValue(viter))) {
    const char *buf2;    
    if (v->type() != LIBERTY__VAL_STRING) {
      char buf[1000];
      sprintf(buf,"%g",v->double_val());
      buf2 = buf;
    } else {
      buf2 = v->string_val();
    }
			
    gs[gsindex-1]->addName(buf2,err);
    
//     if (err == LIB_OBJECT_ALREADY_EXISTS &&
// 	(ge != LIBERTY_GROUPENUM_internal_power)) {
//       sprintf(PB,"%s:%d: The group name %s is already used in this context.",
// 	      LibParse::filename(), lineno, buf2);
//       (*MsgPrinter)(LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, PB);
//     }
    if (err == LIB_INVALID_NAME) {
      sprintf(PB,"%s:%d: The group name \"%s\" is invalid.",
	      LibParse::filename(), lineno, buf2);
      (*MsgPrinter)(LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, PB);
    }
  }
}

void
pop_group()
{
  gsindex--;
}

void
make_complex(LibHead* h)
{
  if (!h || !gs[gsindex-1]) {
    return;
  }
  curr_attr = gs[gsindex-1]->createAttr(h->name(), LIBERTY__COMPLEX,
					h->fileInfo(), err);
  curr_attr->valueList(h->list());
  h->list(NULL);
}

void
make_simple(char *name, LibVal *v)
{
  if (!gs[gsindex-1]) {
    delete v;
    return;
  }
  LibFileInfo f(lineno);
  curr_attr = gs[gsindex-1]->createAttr(name, LIBERTY__SIMPLE, f, err);
  curr_attr->value(v);
}

LibValType
convert_vt(char *type)
{
  if (!strcmp(type,"string")) {
    return LIBERTY__VAL_STRING;
  }
  if (!strcmp(type,"integer")) {
    return LIBERTY__VAL_INT;
  }
  if (!strcmp(type,"float")) {
    return LIBERTY__VAL_DOUBLE;
  }
  return LIBERTY__VAL_UNDEFINED; // to silence g++
}

void
yyerror(const char *s)
{
  //  LibErrorT err;
	
  LibMsgHandlerT MsgPrinter;

  MsgPrinter = LibParse::msgHandler(); /* the printer is in another file! */

  sprintf(PB,"===\nERROR === %s file: %s, line number %d\nERROR ===", s,
	  LibParse::filename(), lineno);
  (*MsgPrinter)(LIB_SEVERITY_ERR, LIB_SYNTAX_ERROR, PB);
	
  syntax_errors++;
}

bool
supported(const char* nm, bool isAttr)
{
  if (isAttr) {
    const libAttrMap* lam = Perfect_Hash_attr::lookup_attr_name(nm, strlen(nm));
    if (!lam) {
      return false;
    }
    switch (lam->type) {
    case LIBERTY_ATTRENUM_index_1                                     :
    case LIBERTY_ATTRENUM_index_2                                     :
    case LIBERTY_ATTRENUM_index_3                                     :
    case LIBERTY_ATTRENUM_values                                      :
    case LIBERTY_ATTRENUM_capacitive_load_unit                        :
    case LIBERTY_ATTRENUM_members                                     :
    case LIBERTY_ATTRENUM_technology                                  :
    case LIBERTY_ATTRENUM_voltage_map                                 :
    case LIBERTY_ATTRENUM_input_voltage_range                         :
    case LIBERTY_ATTRENUM_output_voltage_range                        :
    case LIBERTY_ATTRENUM_rail_connection                             :
    case LIBERTY_ATTRENUM_power_rail                                  :
    case LIBERTY_ATTRENUM_power_gating_pin                            :
    case LIBERTY_ATTRENUM_retention_pin                               :
    case LIBERTY_ATTRENUM_mode                                        :
    case LIBERTY_ATTRENUM_edges                                       :
    case LIBERTY_ATTRENUM_shifts                                      :
      return true;
    default:
      return false;
    }
  } else {
    const libGroupMap* lgm = Perfect_Hash_group::lookup_group_name(nm,strlen(nm));
    if (!lgm) {
      return false;
    }
    switch (lgm->type) {
    case LIBERTY_GROUPENUM_library                     :
    case LIBERTY_GROUPENUM_power_supply                : // ( [power_supply_name] )
    case LIBERTY_GROUPENUM_cell                        : // ( cell_name )
    case LIBERTY_GROUPENUM_scaled_cell                 : // ( scaled_cell_name_1,scaled_cell_name_2 )
    case LIBERTY_GROUPENUM_lu_table_template           : // ( lu_table_template_name )
    case LIBERTY_GROUPENUM_scaling_factors             : // ( scaling_factors_name )
    case LIBERTY_GROUPENUM_operating_conditions        : // ( operating_conditions_name )
    case LIBERTY_GROUPENUM_type                        : // ( type_name )
    case LIBERTY_GROUPENUM_bundle                      : // ( bundle_name )
    case LIBERTY_GROUPENUM_bus                         : // ( bus_name )
    case LIBERTY_GROUPENUM_pin                         :
    case LIBERTY_GROUPENUM_pg_pin                      :
    case LIBERTY_GROUPENUM_ff                          : // ( ff_name_1,ff_name_2 )
    case LIBERTY_GROUPENUM_generated_clock             : // ( generated_clock_name )
    case LIBERTY_GROUPENUM_latch                       : // ( latch_name_1,latch_name_2 )
    case LIBERTY_GROUPENUM_test_cell                   : // ( [test_cell_name] )
    case LIBERTY_GROUPENUM_timing                      : // ( [timing_name] )
    case LIBERTY_GROUPENUM_cell_fall                                   : // ( cell_fall_name )
    case LIBERTY_GROUPENUM_cell_rise                                   : // ( cell_rise_name )
    case LIBERTY_GROUPENUM_fall_constraint                             : // ( fall_constraint_name )
    case LIBERTY_GROUPENUM_fall_propagation                            : // ( fall_propagation_name )
    case LIBERTY_GROUPENUM_fall_transition                             : // ( fall_transition_name )
    case LIBERTY_GROUPENUM_rise_constraint                             : // ( rise_constraint_name )
    case LIBERTY_GROUPENUM_rise_propagation                            : // ( rise_propagation_name )
    case LIBERTY_GROUPENUM_rise_transition                             : // ( rise_transition_name )
    case LIBERTY_GROUPENUM_leakage_power               : // ( [leakage_power_name] )
    case LIBERTY_GROUPENUM_mode_definition             : // ( mode_definition_name )
    case LIBERTY_GROUPENUM_mode_value                  : // ( mode_value_name )
      return true;
    default:
      return false;
    }
  }
  return false;
}

void
deleteParams(LibValEList* paramList)
{
  if (!paramList) {
    return;
  }

  LibVal* val;
  LibValEList::iterator iter = paramList->begin();

  while (iter) {
    val = *iter;
    paramList->remove(iter++);
    delete val;
  }
}
